/*
  PS2Keyboard.cpp - PS2Keyboard library
  Copyright (c) 2007 Free Software Foundation.  All right reserved.
  Written by Christian Weichel <info@32leaves.net>

  ** Mostly rewritten Paul Stoffregen <paul@pjrc.com> 2010, 2011
  ** Modified for use beginning with Arduino 13 by L. Abraham Smith, <n3bah@microcompdesign.com> * 
  ** Modified for easy interrup pin assignement on method begin(datapin,irq_pin). Cuningan <cuninganreset@gmail.com> **
  ** Modified for usage with Sharp MZ-3500 by Martin Lukasek <martin@8bity.cz> 2013

  for more information you can read the original wiki in arduino.cc
  at http://www.arduino.cc/playground/Main/PS2Keyboard
  or http://www.pjrc.com/teensy/td_libs_PS2Keyboard.html

  Version MZ 2.5 (June 2013) - Martin Lukasek
  - Support for Ctrl + ShiftCtrl
  - Support for communication back to keyboard
  - Hardcoded Sharp MZ-3500 communication - no more universal library

  Version 2.4 (March 2013)
  - Support Teensy 3.0, Arduino Due, Arduino Leonardo & other boards
  - French keyboard layout, David Chochoi, tchoyyfr at yahoo dot fr

  Version 2.3 (October 2011)
  - Minor bugs fixed

  Version 2.2 (August 2011)
  - Support non-US keyboards - thanks to Rainer Bruch for a German keyboard :)

  Version 2.1 (May 2011)
  - timeout to recover from misaligned input
  - compatibility with Arduino "new-extension" branch
  - TODO: send function, proposed by Scott Penrose, scooterda at me dot com

  Version 2.0 (June 2010)
  - Buffering added, many scan codes can be captured without data loss
    if your sketch is busy doing other work
  - Shift keys supported, completely rewritten scan code to ascii
  - Slow linear search replaced with fast indexed table lookups
  - Support for Teensy, Arduino Mega, and Sanguino added

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "PS2Keyboard.h"

#define BUFFER_SIZE 45
static volatile uint8_t buffer[BUFFER_SIZE];
static volatile uint8_t head, tail;
static uint8_t DataPin;
static uint8_t IRQ_Pin;
static uint8_t CharBuffer=0;
static uint8_t UTF8next=0;
static const PS2Keymap_t *keymap=NULL;


// ******** KIWI

// vairables used in sending command bytes to the keyboard, eg caps_lock light
volatile boolean cmd_in_progress;
volatile int     cmd_count;
         byte    cmd_value;
volatile byte    cmd_ack_value;
         byte    cmd_parity;
volatile boolean cmd_ack_byte_ok;
volatile byte    kbd_lck_state;

// sending command bytes to the keybaord needs proper parity (otherwise the keyboard
// just asks you to repeat the byte)
byte odd_parity(byte val) {
  int i, count = 1;  // start with 0 for even parity
  for (i=0; i<8; i++) {
    if (val&1) count++;
    val = val>>1;
  }
  return count & 1; // bottom bit of count is parity bit
}

void kbd_send_command(byte val) {
  // stop interrupt routine from receiving characters so that we can use it
  // to send a byte
  cmd_in_progress = true;
  cmd_count       = 0;

  // set up the byte to shift out and initialise the ack bit
  cmd_value      = val;
  cmd_ack_value  = 1;    // the kbd will clear this bit on receiving the byte
  cmd_parity     = odd_parity(val);

  // set the data pin as an output, ready for driving
  digitalWrite(DataPin, HIGH);
  pinMode(DataPin, OUTPUT);

  // drive clock pin low - this is going to generate the first
  // interrupt of the shifting out process
  pinMode(IRQ_Pin, OUTPUT);
  digitalWrite(IRQ_Pin, LOW);

  // wait at least one clock cycle (in case the kbd is mid transmission)
  delayMicroseconds(100);

  // set up the 0 start bit
  digitalWrite(DataPin, LOW);
  // let go of clock - the kbd takes over driving the clock from here
  digitalWrite(IRQ_Pin, HIGH);
  pinMode(IRQ_Pin, INPUT);

  // wait for interrupt routine to shift out byte, parity and receive ack bit
  while (cmd_ack_value!=0) ;
  
  // switch back to the interrupt routine receiving characters from the kbd
  cmd_in_progress = false;

  delay(30);
}

void PS2Keyboard::reset() {
  kbd_send_command(0xFF);   // send the kbd reset code to the kbd: 3 lights
                            // should flash briefly on the kbd

  // reset all the global variables
  /*
  ps2Keyboard_CurrentBuffer = 0;
  ps2Keyboard_CharBuffer    = 0;
  ps2Keyboard_BufferPos     = 0;
  ps2Keyboard_shift         = false;
  ps2Keyboard_ctrl          = false;
  ps2Keyboard_alt           = false;
  ps2Keyboard_extend        = false;
  ps2Keyboard_release       = false;
  ps2Keyboard_caps_lock     = false; */
  cmd_in_progress           = false;
  cmd_count                 = 0;
  cmd_value                 = 0;
  cmd_ack_value             = 1;
  kbd_lck_state             = 0;

  delay(200);
}

// val : bit_2=caps_lock, bit_1=num_lock, bit_0=scroll_lock
void kbd_set_lights(byte val) {
  // When setting the lights with the 0xED command the keyboard responds
  // with an "ack byte", 0xFA. This is NOT the same as the "ack bit" that
  // follows the succesful shifting of each command byte. See this web
  // page for a good description of all this:
  // http://www.beyondlogic.org/keyboard/keybrd.htm
  cmd_ack_byte_ok = false;   // initialise the ack byte flag
  kbd_send_command(0xED);    // send the command byte
  while (!cmd_ack_byte_ok) ; // ack byte from keyboard sets this flag
  // Serial.println(" CMD SENT");
  kbd_send_command(val);     // now send the data
}


// The ISR for the external interrupt
void ps2interrupt(void)
{
	static uint8_t bitcount=0;
	static uint8_t incoming=0;
	static uint32_t prev_ms=0;
	uint32_t now_ms;
	uint8_t n, val;

	val = digitalRead(DataPin);

  // This is the code to send a byte to the keyboard. Actually its 12 bits:
  // a start bit, 8 data bits, 1 parity, 1 stop bit, 1 ack bit (from the kbd)
  if (cmd_in_progress) {
    cmd_count++;          // cmd_count keeps track of the shifting
    switch (cmd_count) {
    case 1: // start bit
      digitalWrite(DataPin,LOW);
      break;
    case 2: case 3: case 4: case 5: case 6: case 7: case 8: case 9:
      // data bits to shift
      digitalWrite(DataPin,cmd_value&1);
      cmd_value = cmd_value>>1;
      break;
    case 10:  // parity bit
      digitalWrite(DataPin,cmd_parity);
      break;
    case 11:  // stop bit
      // release the data pin, so stop bit actually relies on pull-up
      // but this ensures the data pin is ready to be driven by the kbd for
      // for the next bit.
      digitalWrite(DataPin, HIGH);
      pinMode(DataPin, INPUT);
      break;
    case 12: // ack bit - driven by the kbd, so we read its value
      cmd_ack_value = digitalRead(DataPin);
      cmd_in_progress = false;  // done shifting out
    }
    return; // don't fall through to the receive section of the ISR
  } else {



	now_ms = millis();
	if (now_ms - prev_ms > 250) {
		bitcount = 0;
		incoming = 0;
	}
	prev_ms = now_ms;
	n = bitcount - 1;
	if (n <= 7) {
		incoming |= (val << n);
	}
	bitcount++;
	if (bitcount == 11) {
                if( incoming == 0xFA ) { // ACK From keyboard
                  cmd_ack_byte_ok = true;
                }
		uint8_t i = head + 1;
		if (i >= BUFFER_SIZE) i = 0;
		if (i != tail) {
			buffer[i] = incoming;
			head = i;
		}
		bitcount = 0;
		incoming = 0;
	}
  }
}

static inline uint8_t get_scan_code(void)
{
	uint8_t c, i;

	i = tail;
	if (i == head) return 0;
	i++;
	if (i >= BUFFER_SIZE) i = 0;
	c = buffer[i];
	tail = i;
	return c;
}

// http://www.quadibloc.com/comp/scan.htm
// http://www.computer-engineering.org/ps2keyboard/scancodes2.html

// These arrays provide a simple key map, to turn scan codes into ISO8859
// output.  If a non-US keyboard is used, these may need to be modified
// for the desired output.
//

const PROGMEM PS2Keymap_t PS2Keymap_US = {
  // without shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, '`', 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'q', '1', 0,
	0, 0, 'z', 's', 'a', 'w', '2', 0,
	0, 'c', 'x', 'd', 'e', '4', '3', 0,
	0, ' ', 'v', 'f', 't', 'r', '5', 0,
	0, 'n', 'b', 'h', 'g', 'y', '6', 0,
	0, 0, 'm', 'j', 'u', '7', '8', 0,
	0, ',', 'k', 'i', 'o', '0', '9', 0,
	0, '.', '/', 'l', ';', 'p', '-', 0,
	0, 0, '\'', 0, '[', '=', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, ']', 0, '\\', 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
  // with shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, '~', 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'Q', '!', 0,
	0, 0, 'Z', 'S', 'A', 'W', '@', 0,
	0, 'C', 'X', 'D', 'E', '$', '#', 0,
	0, ' ', 'V', 'F', 'T', 'R', '%', 0,
	0, 'N', 'B', 'H', 'G', 'Y', '^', 0,
	0, 0, 'M', 'J', 'U', '&', '*', 0,
	0, '<', 'K', 'I', 'O', ')', '(', 0,
	0, '>', '?', 'L', ':', 'P', '_', 0,
	0, 0, '"', 0, '{', '+', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '}', 0, '|', 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
	0, 0
};


const PROGMEM PS2Keymap_t PS2Keymap_German = {
  // without shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, '^', 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'q', '1', 0,
	0, 0, 'y', 's', 'a', 'w', '2', 0,
	0, 'c', 'x', 'd', 'e', '4', '3', 0,
	0, ' ', 'v', 'f', 't', 'r', '5', 0,
	0, 'n', 'b', 'h', 'g', 'z', '6', 0,
	0, 0, 'm', 'j', 'u', '7', '8', 0,
	0, ',', 'k', 'i', 'o', '0', '9', 0,
	0, '.', '-', 'l', PS2_o_DIAERESIS, 'p', PS2_SHARP_S, 0,
	0, 0, PS2_a_DIAERESIS, 0, PS2_u_DIAERESIS, '\'', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '+', 0, '#', 0, 0,
	0, '<', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
  // with shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, PS2_DEGREE_SIGN, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'Q', '!', 0,
	0, 0, 'Y', 'S', 'A', 'W', '"', 0,
	0, 'C', 'X', 'D', 'E', '$', PS2_SECTION_SIGN, 0,
	0, ' ', 'V', 'F', 'T', 'R', '%', 0,
	0, 'N', 'B', 'H', 'G', 'Z', '&', 0,
	0, 0, 'M', 'J', 'U', '/', '(', 0,
	0, ';', 'K', 'I', 'O', '=', ')', 0,
	0, ':', '_', 'L', PS2_O_DIAERESIS, 'P', '?', 0,
	0, 0, PS2_A_DIAERESIS, 0, PS2_U_DIAERESIS, '`', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '*', 0, '\'', 0, 0,
	0, '>', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
	1, 0,
  // with altgr
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, 0, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, '@', 0, 0,
	0, 0, 0, 0, 0, 0, PS2_SUPERSCRIPT_TWO, 0,
	0, 0, 0, 0, PS2_CURRENCY_SIGN, 0, PS2_SUPERSCRIPT_THREE, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, PS2_MICRO_SIGN, 0, 0, '{', '[', 0,
	0, 0, 0, 0, 0, '}', ']', 0,
	0, 0, 0, 0, 0, 0, '\\', 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '~', 0, '#', 0, 0,
	0, '|', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 }
};


const PROGMEM PS2Keymap_t PS2Keymap_French = {
  // without shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, PS2_SUPERSCRIPT_TWO, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'a', '&', 0,
	0, 0, 'w', 's', 'q', 'z', PS2_e_ACUTE, 0,
	0, 'c', 'x', 'd', 'e', '\'', '"', 0,
	0, ' ', 'v', 'f', 't', 'r', '(', 0,
	0, 'n', 'b', 'h', 'g', 'y', '-', 0,
	0, 0, ',', 'j', 'u', PS2_e_GRAVE, '_', 0,
	0, ';', 'k', 'i', 'o', PS2_a_GRAVE, PS2_c_CEDILLA, 0,
	0, ':', '!', 'l', 'm', 'p', ')', 0,
	0, 0, PS2_u_GRAVE, 0, '^', '=', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '$', 0, '*', 0, 0,
	0, '<', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
  // with shift
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, 0, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'A', '1', 0,
	0, 0, 'W', 'S', 'Q', 'Z', '2', 0,
	0, 'C', 'X', 'D', 'E', '4', '3', 0,
	0, ' ', 'V', 'F', 'T', 'R', '5', 0,
	0, 'N', 'B', 'H', 'G', 'Y', '6', 0,
	0, 0, '?', 'J', 'U', '7', '8', 0,
	0, '.', 'K', 'I', 'O', '0', '9', 0,
	0, '/', PS2_SECTION_SIGN, 'L', 'M', 'P', PS2_DEGREE_SIGN, 0,
	0, 0, '%', 0, PS2_DIAERESIS, '+', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, PS2_POUND_SIGN, 0, PS2_MICRO_SIGN, 0, 0,
	0, '>', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 },
	1, 0,
  // with altgr
	{0, PS2_F9, 0, PS2_F5, PS2_F3, PS2_F1, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, 0, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, '@', 0, 0,
	0, 0, 0, 0, 0, 0, '~', 0,
	0, 0, 0, 0, 0 /*PS2_EURO_SIGN*/, '{', '#', 0,
	0, 0, 0, 0, 0, 0, '[', 0,
	0, 0, 0, 0, 0, 0, '|', 0,
	0, 0, 0, 0, 0, '`', '\\', 0,
	0, 0, 0, 0, 0, '@', '^', 0,
	0, 0, 0, 0, 0, 0, ']', 0,
	0, 0, 0, 0, 0, 0, '}', 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '¤', 0, '#', 0, 0,
	0, '|', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	PS2_F11, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, PS2_F7 }
};

const PROGMEM PS2Keymap_t PS2Keymap_MZ3500 = {
  // without shift
	{0, 0x09/*F9*/, 0, 0x05/*PS2_F5*/, 0x03/*PS2_F3*/, 0x01/*PS2_F1*/, 0x02/*PS2_F2*/, 0xDB /*PS2_F12*/,
	0, 0x0A/*PS2_F10*/, 0x08/*PS2_F8*/, 0x06/*PS2_F6*/, 0x04/*PS2_F4*/, 0xD4/*PS2_TAB*/, '`', 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'Q', '1', 0,
	0, 0, 'Z', 'S', 'A', 'W', '2', 0,
	0, 'C', 'X', 'D', 'E', '4', '3', 0,
	0, ' ', 'V', 'F', 'T', 'R', '5', 0,
	0, 'N', 'B', 'H', 'G', 'Y', '6', 0,
	0, 0, 'M', 'J', 'U', '7', '8', 0,
	0, ',', 'K', 'I', 'O', '0', '9', 0,
	0, '.', '/', 'L', ';', 'P', '-', 0,
	0, 0, '\'', 0, '[', '=', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, ']', 0, '\\', 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', 0xFD, '2', '5', '6', '8', 0x0F, 0 /*NumLock*/,
	0xD5 /*PS2_F11*/, '+', '3', 0x0E/*'-'*/, '*', '9', PS2_SCROLL, 0,
	0, 0, 0, 0x07/*PS2_F7*/ },
  // with shift
	{0, 0x09/*F9*/, 0, 0x05/*PS2_F5*/, 0x03/*PS2_F3*/, 0x01/*PS2_F1*/, 0x02/*PS2_F2*/, 0xDB /*PS2_F12*/,
	0, 0x0A/*PS2_F10*/, 0x08/*PS2_F8*/, 0x06/*PS2_F6*/, 0x04/*PS2_F4*/, 0xD4/*PS2_TAB*/, '~', 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 'q', '!', 0,
	0, 0, 'z', 's', 'a', 'w', '@', 0,
	0, 'c', 'x', 'd', 'e', '$', '#', 0,
	0, ' ', 'v', 'f', 't', 'r', '%', 0,
	0, 'n', 'b', 'h', 'g', 'y', '^', 0,
	0, 0, 'm', 'j', 'u', '&', '*', 0,
	0, '<', 'k', 'i', 'o', ')', '(', 0,
	0, '>', '?', 'l', ':', 'p', '_', 0,
	0, 0, '"', 0, '{', '+', 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '}', 0, '|', 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	'0', 0xFD, '2', '5', '6', '8', 0x0F, 0 /*NumLock*/,
	0xD5 /*PS2_F11*/, '+', '3', 0x0E/*'-'*/, '*', '9', PS2_SCROLL, 0,
	0, 0, 0, 0x07/*PS2_F7*/ },
	1, 1,
  // with altgr
	{0, PS2_F9, 0, PS2_F5, PS2_F3, 0xDA /*PS2_F1*/, PS2_F2, PS2_F12,
	0, PS2_F10, PS2_F8, PS2_F6, PS2_F4, PS2_TAB, 0, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, '@', 0, 0,
	0, 0, 0, 0, 0, 0, '~', 0, 
	0, 0, 0, 0, 0 /*PS2_EURO_SIGN*/, '{', '#', 0,
	0, 0, 0, 0, 0, 0, '[', 0,
	0, 0, 0, 0, 0, 0, '|', 0,
	0, 0, 0, 0, 0, '`', '\\', 0,
	0, 0, 0, 0, 0, '@', '^', 0,
	0, 0, 0, 0, 0, 0, ']', 0,
	0, 0, 0, 0, 0, 0, '}', 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, '¤', 0, '#', 0, 0,
	0, '|', 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, '1', 0, '4', '7', 0, 0, 0,
	0xFC/*'0'->'00'*/, '.', '2', '5', '6', '8', PS2_ESC, 0 /*NumLock*/,
	0/*PS2_F11*/, '+', '3', '-', '*', '9', PS2_SCROLL, 0,
	0, 0, 0, 0x07/*PS2_F7*/ },
  // with ctrl
	{0, 0x19/*F9*/, 0, 0x15/*PS2_F5*/, 0x13/*PS2_F3*/, 0x11/*PS2_F1*/, 0x12/*PS2_F2*/, 0xD3 /*PS2_F12*/,
	0, 0x1A/*PS2_F10*/, 0x18/*PS2_F8*/, 0x16/*PS2_F6*/, 0x14/*PS2_F4*/, 0xD4/*PS2_TAB*/, 0x9D, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 0x91, 0, 0,
	0, 0, 0x9A, 0x93, 0x81, 0x97, 0, 0,
	0, 0x83, 0x98, 0x84, 0x85, 0, 0, 0,
	0, ' ', 0x96, 0x86, 0x94, 0x92, 0, 0,
	0, 0x8E, 0x82, 0x88, 0x87, 0x99, 0, 0,
	0, 0, 0x8D, 0x8A, 0x95, 0, 0, 0,
	0, ',', 0x8B, 0x89, 0x8F, 0, 0, 0,
	0, '.', '/', 0x8C, ';', 0x90, 0 /*'-'*/, 0,
	0, 0, 0x9F /*'\''*/, 0, 0x80 /*'['*/, 0x9E/*'='*/, 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, 0x9B /*']'*/, 0, 0x9C /*'\\'*/, 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, 0xA1 /*'1'*/, 0, 0xA4 /*'4'*/, 0xA7 /*'7'*/, 0, 0, 0,
	0xA0 /*'0'*/, 0xFD, 0xA2 /*'2'*/, 0xA5 /*'5'*/, 0xA6 /*'6'*/, 0xA8 /*'8'*/, 0x0F, 0 /*NumLock*/,
	0xD6 /*PS2_F11*/, '+', 0xA3 /*'3'*/, 0x0E/*'-'*/, '*', 0xA9 /*'9'*/, PS2_SCROLL, 0,
	0, 0, 0, 0x17/*PS2_F7*/ },
  // with shift ctrl
	{0, 0x19/*F9*/, 0, 0x15/*PS2_F5*/, 0x13/*PS2_F3*/, 0x11/*PS2_F1*/, 0x12/*PS2_F2*/, 0xD3 /*PS2_F12*/,
	0, 0x1A/*PS2_F10*/, 0x18/*PS2_F8*/, 0x16/*PS2_F6*/, 0x14/*PS2_F4*/, 0xD4/*PS2_TAB*/, 0, 0,
	0, 0 /*Lalt*/, 0 /*Lshift*/, 0, 0 /*Lctrl*/, 0xF1, 0, 0,
	0, 0, 0xFA, 0xF3, 0xE1, 0xF7, 0, 0,
	0, 0xE3, 0xF8, 0xE4, 0xE5, 0, 0, 0,
	0, ' ', 0xF6, 0xE6, 0xF4, 0xF2, 0, 0,
	0, 0xEE, 0xE2, 0xE8, 0xE7, 0xF9, 0, 0,
	0, 0, 0xED, 0xEA, 0xF5, 0, 0, 0,
	0, ',', 0xEB, 0xE9, 0xEF, 0, 0, 0,
	0, '.', '/', 0xEC, ';', 0xF0, 0/*'-'*/, 0,
	0, 0, 0 /*'\''*/, 0, 0xE0 /*'['*/, 0 /*'='*/, 0, 0,
	0 /*CapsLock*/, 0 /*Rshift*/, PS2_ENTER /*Enter*/, 0xFB/*']'*/, 0, 0 /*'\\'*/, 0, 0,
	0, 0, 0, 0, 0, 0, PS2_BACKSPACE, 0,
	0, 0xA1 /*'1'*/, 0, 0xA4 /*'4'*/, 0xA7 /*'7'*/, 0, 0, 0,
	0 /*'0'*/, 0xFD, 0xA2 /*'2'*/, 0xA5 /*'5'*/, 0xA6 /*'6'*/, 0xA8 /*'8'*/, 0x0F, 0 /*NumLock*/,
	0xD6 /*PS2_F11*/, '+', 0xA3 /*'3'*/, 0x0E/*'-'*/, '*', 0xA9 /*'9'*/, PS2_SCROLL, 0,
	0, 0, 0, 0x17/*PS2_F7*/ }
};

#define BREAK     0x01
#define MODIFIER  0x02
#define SHIFT_L   0x04
#define SHIFT_R   0x08
#define ALTGR     0x10
#define CTRL      0x20

char get_iso8859_code(void)
{
	static uint8_t state=0;
	uint8_t s;
	char c;

	while (1) {
		s = get_scan_code();
		if (!s) return 0;
                // Serial.print("-"); Serial.print(byte(s)); Serial.print("-");
		if (s == 0xF0) {
			state |= BREAK;
		} else if (s == 0xE0) {
			state |= MODIFIER;
		} else {
			if (state & BREAK) {
				if (s == 0x12) {
					state &= ~SHIFT_L;
				} else if (s == 0x59) {
					state &= ~SHIFT_R;
				} else if (s == 0x11) { // (s == 0x11 && (state & MODIFIER))
					state &= ~ALTGR;
				} else if (s == 0x14) { // (s == 0x11 && (state & MODIFIER))
					state &= ~CTRL;
				}
				// CTRL, ALT & WIN keys could be added
				// but is that really worth the overhead?
				state &= ~(BREAK | MODIFIER);
				continue;
			}
			if (s == 0x12) {
				state |= SHIFT_L;
				continue;
			} else if (s == 0x59) {
				state |= SHIFT_R;
				continue;
			} else if (s == 0x11) { // (s == 0x11 && (state & MODIFIER))
				state |= ALTGR;
				continue;
			} else if (s == 0x14) { // (s == 0x11 && (state & MODIFIER))
				state |= CTRL;
				continue;
			} else if (s == 0x58) { // Caps Lock
                                if (kbd_lck_state & LED_CAPSLOCK) {
                                  kbd_lck_state &= ~LED_CAPSLOCK;
                                } else {
                                  kbd_lck_state |= LED_CAPSLOCK;
                                }
                                kbd_set_lights(kbd_lck_state);
				continue;
			} else if (s == 0x77) { // Num Lock
                                /* always numlock
                                if (kbd_lck_state & LED_NUMLOCK) {
                                  kbd_lck_state &= ~LED_NUMLOCK;
                                } else {
                                  kbd_lck_state |= LED_NUMLOCK;
                                } */
                                kbd_set_lights(kbd_lck_state);
				continue;
			} else if (s == 0x7E) { // Scroll Lock
                                if (kbd_lck_state & LED_SCRLLOCK) {
                                  kbd_lck_state &= ~LED_SCRLLOCK;
                                } else {
                                  kbd_lck_state |= LED_SCRLLOCK;
                                }
                                kbd_set_lights(kbd_lck_state);
				// continue;
                                c = 0xD0;
                                return c;
			}
			c = 0;
			if (state & MODIFIER) {
				switch (s) {
				  case 0x70: c = 0xD9;            break; // PS2_INSERT
				  case 0x71: c = 0xD8;            break; // PS2_DELETE
				  case 0x6C: 
                                    c = 0x0B;  // HOME
                                    if ((s < PS2_KEYMAP_SIZE) && ((state & CTRL) && keymap->uses_ctrl))
                                      c = 0x1B; // Ctrl HOME = ESC
                                    break;
				  case 0x69: c = 0xD7;            break; // PS2_END
				  case 0x7D: c = PS2_PAGEUP;      break;
				  case 0x7A: c = PS2_PAGEDOWN;    break;
				  case 0x75: c = 0x1C;            break; // PS2_UPARROW
				  case 0x6B: c = 0x1F;            break; // PS2_LEFTARROW
				  case 0x72: c = 0x1D;            break; // PS2_DOWNARROW
				  case 0x74: c = 0x1E;            break; // PS2_RIGHTARROW
				  case 0x4A: c = '/';             break; // NUM /
				  case 0x5A: c = PS2_ENTER;       break; // NUM Enter
				  case 0x1F:                             // Left Windows Key
                                  case 0x27:                             // Rightt Windows Key 
                                    c = 0xFE;  // Windows -> CMD
                                    if ((s < PS2_KEYMAP_SIZE) && ((state & CTRL) && keymap->uses_ctrl))
                                      c = 0xDC; // Ctrl Windows -> Ctrl CMD
                                    break;
				  default: break;
				}
			} else if ((state & ALTGR) && keymap->uses_altgr) {
				if (s < PS2_KEYMAP_SIZE)
					c = pgm_read_byte(keymap->altgr + s);
			} else if ((state & (SHIFT_L | SHIFT_R)) && (state & CTRL) && keymap->uses_ctrl) {
				if (s < PS2_KEYMAP_SIZE)
                                  if (kbd_lck_state & LED_CAPSLOCK) {
					c = pgm_read_byte(keymap->ctrl + s);
                                  } else {
					c = pgm_read_byte(keymap->shiftctrl + s);
                                  }
			} else if ((state & CTRL) && keymap->uses_ctrl) {
				if (s < PS2_KEYMAP_SIZE)
                                  if (kbd_lck_state & LED_CAPSLOCK) {
					c = pgm_read_byte(keymap->shiftctrl + s);
                                  } else {
					c = pgm_read_byte(keymap->ctrl + s);
                                  }
			} else if ((state & (SHIFT_L | SHIFT_R)) && !(state & CTRL)) {
				if (s < PS2_KEYMAP_SIZE) {
                                  if (kbd_lck_state & LED_CAPSLOCK) {
					c = pgm_read_byte(keymap->noshift + s);
                                  } else {
					c = pgm_read_byte(keymap->shift + s);
                                  }
                                }
			} else {
				if (s < PS2_KEYMAP_SIZE)
                                  if (kbd_lck_state & LED_CAPSLOCK) {
					c = pgm_read_byte(keymap->shift + s);
                                  } else {
					c = pgm_read_byte(keymap->noshift + s);
                                  }
			}
			state &= ~(BREAK | MODIFIER);
			if (c) return c;
		}
	}
}

bool PS2Keyboard::available() {
	if (CharBuffer || UTF8next) return true;
	CharBuffer = get_iso8859_code();
	if (CharBuffer) return true;
	return false;
}

int PS2Keyboard::read() {
	uint8_t result;

	result = UTF8next;
	if (result) {
		UTF8next = 0;
	} else {
		result = CharBuffer;
		if (result) {
			CharBuffer = 0;
		} else {
			result = get_iso8859_code();
		}
/*		if (result >= 128) {
			UTF8next = (result & 0x3F) | 0x80;
			result = ((result >> 6) & 0x1F) | 0xC0;
		} */
	}
	if (!result) return -1;
	return result;
}

PS2Keyboard::PS2Keyboard() {
  // nothing to do here, begin() does it all
}

void PS2Keyboard::begin(uint8_t data_pin, uint8_t irq_pin, const PS2Keymap_t &map) {
  uint8_t irq_num=255;

  DataPin = data_pin;
  IRQ_Pin = irq_pin;
  keymap = &map;

  // initialize the pins
#ifdef INPUT_PULLUP
  pinMode(irq_pin, INPUT_PULLUP);
  pinMode(data_pin, INPUT_PULLUP);
#else
  pinMode(irq_pin, INPUT);
  digitalWrite(irq_pin, HIGH);
  pinMode(data_pin, INPUT);
  digitalWrite(data_pin, HIGH);
#endif

#ifdef CORE_INT_EVERY_PIN
  irq_num = irq_pin;

#else
  switch(irq_pin) {
    #ifdef CORE_INT0_PIN
    case CORE_INT0_PIN:
      irq_num = 0;
      break;
    #endif
    #ifdef CORE_INT1_PIN
    case CORE_INT1_PIN:
      irq_num = 1;
      break;
    #endif
    #ifdef CORE_INT2_PIN
    case CORE_INT2_PIN:
      irq_num = 2;
      break;
    #endif
    #ifdef CORE_INT3_PIN
    case CORE_INT3_PIN:
      irq_num = 3;
      break;
    #endif
    #ifdef CORE_INT4_PIN
    case CORE_INT4_PIN:
      irq_num = 4;
      break;
    #endif
    #ifdef CORE_INT5_PIN
    case CORE_INT5_PIN:
      irq_num = 5;
      break;
    #endif
    #ifdef CORE_INT6_PIN
    case CORE_INT6_PIN:
      irq_num = 6;
      break;
    #endif
    #ifdef CORE_INT7_PIN
    case CORE_INT7_PIN:
      irq_num = 7;
      break;
    #endif
  }
#endif

  cmd_in_progress           = false;
  cmd_count                 = 0;
  cmd_value                 = 0;
  cmd_ack_value             = 1;
  
  head = 0;
  tail = 0;
  if (irq_num < 255) {
    // Serial.print(irq_num);
    attachInterrupt(irq_num, ps2interrupt, FALLING);
  }

}


